//
//  ARView.m
//  ARApp2
//
//  Copyright 2010-2011 ARToolworks, Inc. All rights reserved.
//  Author: Philip Lamb
//

#import <QuartzCore/QuartzCore.h>
#import "ARView.h"
#import "ARViewController.h"
#include "glStateCache.h"
#include <AR/gsub_es.h>
#include <Eden/EdenMath.h>

NSString *const ARViewUpdatedCameraLensNotification = @"ARViewUpdatedCameraLensNotification";
NSString *const ARViewUpdatedCameraPoseNotification = @"ARViewUpdatedCameraPoseNotification";
NSString *const ARViewUpdatedViewportNotification = @"ARViewUpdatedViewportNotification";
NSString *const ARViewDrawPreCameraNotification = @"ARViewDrawPreCameraNotification";
NSString *const ARViewDrawPostCameraNotification = @"ARViewDrawPostCameraNotification";
NSString *const ARViewDrawOverlayNotification = @"ARViewDrawOverlayNotification";

@implementation ARView

@synthesize contentWidth, contentHeight, contentRotate90, touchDelegate, arViewController;

- (id) initWithFrame:(CGRect)frame pixelFormat:(NSString*)format depthFormat:(GLuint)depth stencilFormat:(GLuint)stencil preserveBackbuffer:(BOOL)retained
{
    if ((self = [super initWithFrame:frame pixelFormat:format depthFormat:depth stencilFormat:stencil preserveBackbuffer:retained])) {
         
        // Init instance variables.
        arViewController = nil;
        contentRotate90 = NO;
        contentWidth = (int)frame.size.width;
        contentHeight = (int)frame.size.height;
        cameraPoseValid = NO;
        
        // Init gestures.
        [self setMultipleTouchEnabled:YES];
        //[self setTouchDelegate:self];

        // One-time OpenGL setup goes here.
        glStateCacheBeginAgain();
        
        //CHECK_GL_ERROR();
    }
    
    return (self);
    
}

- (void)layoutSubviews
{
    [super layoutSubviews];
    
    // Calculate viewport.
    int left, bottom, w, h;
    
    int backingWidth = (int)surfaceSize.width;
    int backingHeight = (int)surfaceSize.height;
#ifdef DEBUG
    NSLog(@"[ARView layoutSubviews] backingWidth=%d, backingHeight=%d\n", backingWidth, backingHeight);
#endif        

    if (self.contentMode == UIViewContentModeScaleToFill) {
        w = backingWidth;
        h = backingHeight;
    } else {
        int contentWidthFinalOrientation = (contentRotate90 ? contentHeight : contentWidth);
        int contentHeightFinalOrientation = (contentRotate90 ? contentWidth : contentHeight);
        if (self.contentMode == UIViewContentModeScaleAspectFit || self.contentMode == UIViewContentModeScaleAspectFill) {
            float scaleRatioWidth, scaleRatioHeight, scaleRatio;
            scaleRatioWidth = (float)backingWidth / (float)contentWidthFinalOrientation;
            scaleRatioHeight = (float)backingHeight / (float)contentHeightFinalOrientation;
            if (self.contentMode == UIViewContentModeScaleAspectFill) scaleRatio = MAX(scaleRatioHeight, scaleRatioWidth);
            else scaleRatio = MIN(scaleRatioHeight, scaleRatioWidth);
            w = (int)((float)contentWidthFinalOrientation * scaleRatio);
            h = (int)((float)contentHeightFinalOrientation * scaleRatio);
        } else {
            w = contentWidthFinalOrientation;
            h = contentHeightFinalOrientation;
        }
    }
    
    if (self.contentMode == UIViewContentModeTopLeft
        || self.contentMode == UIViewContentModeLeft
        || self.contentMode == UIViewContentModeBottomLeft) left = 0;
    else if (self.contentMode == UIViewContentModeTopRight
             || self.contentMode == UIViewContentModeRight
             || self.contentMode == UIViewContentModeBottomRight) left = backingWidth - w;
    else left = (backingWidth - w) / 2;
        
    if (self.contentMode == UIViewContentModeBottomLeft
        || self.contentMode == UIViewContentModeBottom
        || self.contentMode == UIViewContentModeBottomRight) bottom = 0;
    else if (self.contentMode == UIViewContentModeTopLeft
             || self.contentMode == UIViewContentModeTop
             || self.contentMode == UIViewContentModeTopRight) bottom = backingHeight - h;
    else bottom = (backingHeight - h) / 2;

    glViewport(left, bottom, w, h);
    
    viewPort[viewPortIndexLeft] = left;
    viewPort[viewPortIndexBottom] = bottom;
    viewPort[viewPortIndexWidth] = w;
    viewPort[viewPortIndexHeight] = h;
    [[NSNotificationCenter defaultCenter] postNotificationName:ARViewUpdatedViewportNotification object:self];
#ifdef DEBUG
    NSLog(@"[ARView layoutSubviews] viewport left=%d, bottom=%d, width=%d, height=%d\n", left, bottom, w, h);
#endif
}

- (GLint *)viewPort
{
    return (viewPort);
}

- (void)setCameraLens:(float *)lens
{
    if (lens) {
        int i;
        for (i = 0; i < 16; i++) cameraLens[i] = lens[i];
        [[NSNotificationCenter defaultCenter] postNotificationName:ARViewUpdatedCameraLensNotification object:self];
    }
}

- (float *)cameraLens
{
    return (cameraLens);
}

- (void)setCameraPose:(float *)pose
{
    if (pose) {
        int i;
        for (i = 0; i < 16; i++) cameraPose[i] = pose[i];
        [[NSNotificationCenter defaultCenter] postNotificationName:ARViewUpdatedCameraPoseNotification object:self];
        cameraPoseValid = TRUE;
    } else {
        cameraPoseValid = FALSE;
    }
}

- (float *)cameraPose
{
    if (cameraPoseValid) return (cameraPose);
    else return (NULL);
}

- (void) drawView
{
	float width, height;
	
    [self setCurrentContext];
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, frameBuffer);
    
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); // Clear the render buffers for new frame.

    arglDispImage(arViewController.arglContextSettings);
    
    // Set up 3D mode.
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if (contentRotate90) glRotatef(90.0f, 0.0f, 0.0f, -1.0f);
    glMultMatrixf(cameraLens);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glStateCacheEnableDepthTest();
    
    // Set any initial per-frame GL state you require here.
    // --->

    // Lighting and geometry that moves with the camera should be added here.
    // (I.e. should be specified before camera pose transform.)
    // --->
    [[NSNotificationCenter defaultCenter] postNotificationName:ARViewDrawPreCameraNotification object:self];
    
    if (cameraPoseValid) {
        
        glMultMatrixf(cameraPose);
        
        // All lighting and geometry to be drawn in world coordinates goes here.
        // --->
        [[NSNotificationCenter defaultCenter] postNotificationName:ARViewDrawPostCameraNotification object:self];
    }
    
    // If you added external OpenGL code above, and that code doesn't use the glStateCache routines,
    // then uncomment the line below.
    //glStateCacheBeginAgain();
    
    // Set up 2D mode.
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if (contentRotate90) glRotatef(90.0f, 0.0f, 0.0f, -1.0f);
	width = (float)viewPort[(contentRotate90 ? viewPortIndexHeight : viewPortIndexWidth)];
	height = (float)viewPort[(contentRotate90 ? viewPortIndexWidth : viewPortIndexHeight)];
	glOrthof(0.0f, width, 0.0f, height, -1.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glStateCacheDisableDepthTest();
    
    // Add your own 2D overlays here.
    // --->
    [[NSNotificationCenter defaultCenter] postNotificationName:ARViewDrawOverlayNotification object:self];
   
    // If you added external OpenGL code above, and that code doesn't use the glStateCache routines,
    // then uncomment the line below.
    //glStateCacheBeginAgain();

#ifdef DEBUG
    // Example of 2D drawing. It just draws a white border line. Change the 0 to 1 to enable.
    const GLfloat square_vertices [4][2] = { {0.5f, 0.5f}, {0.5f, height - 0.5f}, {width - 0.5f, height - 0.5f}, {width - 0.5f, 0.5f} };
    glStateCacheDisableLighting();
    glStateCacheDisableTex2D();
    glStateCacheVertexPtr(2, GL_FLOAT, 0, square_vertices);
    glStateCacheEnableClientStateVertexArray();
    glColor4ub(255, 255, 255, 255);
    glDrawArrays(GL_LINE_LOOP, 0, 4);
#endif
    
#ifdef DEBUG
    CHECK_GL_ERROR();
#endif
        
    [self swapBuffers];
}

- (void) dealloc
{
    [super dealloc];
}

// Handles the start of a touch
- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
    [self.touchDelegate tapDetectingView:self touchesBegan:touches withEvent:event];
    NSLog(@"Test");
}

// Handles the continuation of a touch.
- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
    NSArray*        array = [touches allObjects];
    UITouch*        touch;
    NSUInteger        i;
    
#ifdef DEBUG
    //NSLog(@"[EAGLView touchesMoved].\n");
#endif
    
    for (i = 0; i < [array count]; ++i) {
        touch = [array objectAtIndex:i];
        if (touch.phase == UITouchPhaseMoved) {
            // Can do something appropriate for a moving touch here.
         } // phase match.
    } // touches.
}

// Handles the end of a touch event.
- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
    NSArray*        array = [touches allObjects];
    UITouch*        touch;
    NSUInteger        i;
    
#ifdef DEBUG
    //NSLog(@"[EAGLView touchesEnded].\n");
#endif
    
    for (i = 0; i < [array count]; ++i) {
        touch = [array objectAtIndex:i];
        if (touch.phase == UITouchPhaseEnded) {
            // Can do something appropriate for end of touch here.
        } // phase match.
    } // touches.
}

- (void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event
{
    NSArray*        array = [touches allObjects];
    UITouch*        touch;
    NSUInteger        i;
    
#ifdef DEBUG
    //NSLog(@"[EAGLView touchesCancelled].\n");
#endif
    
    for (i = 0; i < [array count]; ++i) {
        touch = [array objectAtIndex:i];
        if (touch.phase == UITouchPhaseCancelled) {
               // Can do something appropriate for cancellation of a touch (e.g. by a system event) here.
        } // phase match.
    } // touches.
}

- (void)convertPointInViewToRay:(CGPoint)point
{
    
    float m[16], A[16];
    float p[4], q[4];
    
    // Find INVERSE(PROJECTION * MODELVIEW).
    if (contentRotate90) {
        EdenMathRotationMatrix(A, DTOR*90.0f, 0.0f, 0.0f, -1.0f);
        EdenMathMultMatrix(m, cameraLens, A);
        EdenMathMultMatrix(A, m, cameraPose);
    } else {
        EdenMathMultMatrix(A, cameraLens, cameraPose);
    }
    if (EdenMathInvertMatrix4x4(m, A)) {
        
        // Next, normalise point to viewport range [-1.0, 1.0], and with depth -1.0 (i.e. at near clipping plane).
        p[0] = (point.x - viewPort[viewPortIndexLeft]) * 2.0f / viewPort[viewPortIndexWidth] - 1.0f; // (winx - viewport[0]) * 2 / viewport[2] - 1.0;
        p[1] = (point.y - viewPort[viewPortIndexBottom]) * 2.0f / viewPort[viewPortIndexHeight] - 1.0f; // (winy - viewport[1]) * 2 / viewport[3] - 1.0;
        p[2] = -1.0f; // 2 * winz - 1.0;
        p[3] = 1.0f;
        
        // Calculate the point's world coordinates.
        EdenMathMultMatrix4x4ByVector(q, m, p);
        
        if (q[3] != 0.0f) {
            rayPoint1[0] = q[0] / q[3];
            rayPoint1[1] = q[1] / q[3];
            rayPoint1[2] = q[2] / q[3];
            
            // Next, a second point with depth 1.0 (i.e. at far clipping plane).
            p[2] = 1.0f; // 2 * winz - 1.0;
            
            // Calculate the point's world coordinates.
            EdenMathMultMatrix4x4ByVector(q, m, p);
            if (q[3] != 0.0f) {
                
                rayPoint2[0] = q[0] / q[3];
                rayPoint2[1] = q[1] / q[3];
                rayPoint2[2] = q[2] / q[3];
                
                rayIsValid = TRUE;
                return;
            }
        }
    }
    rayIsValid = FALSE;
}

- (void) handleTouchAtLocation:(CGPoint)location tapCount:(NSUInteger)tapCount
{
    float abc[3];
    float d;
    // A plane which lies on the z axis.
    const float p1[3] = {0.0f, 0.0f, 0.0f};
    const float p2[3] = {1.0f, 0.0f, 0.0f};
    const float p3[3] = {0.0f, 1.0f, 0.0f};
    float intersection[3];
    
    CGPoint locationFlippedY = CGPointMake(location.x, surfaceSize.height - location.y);
    
    [self convertPointInViewToRay:locationFlippedY];
    if (rayIsValid) {
        EdenMathPointsToPlane(abc, &d, p1, p2, p3);
        if (EdenMathIntersectionLinePlane(intersection, NULL, rayPoint1, rayPoint2, abc, d)) { // Test if the ray intersects the z-plane (defined above).
            
            // If it does, do something trivial.
            NSLog(@"Touch at (%f,%f).\n", intersection[0], intersection[1]);
        }
    }
}

@end
